class LMFeatureExtractor(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, seq1h, LM_output):
        seqlen = seq1h.size(1)-2
        
        features_1d = LM_output['representations'][33][:,1:-1,:]
        features_1d = torch.cat([seq1h[:,1:-1], features_1d], axis=-1)

        features_2d = LM_output['attentions'][:,:,:,1:-1,1:-1] # 1, 33, 20, 110, 110
        features_2d = features_2d.reshape(-1, 660, seqlen, seqlen)
        features_2d = torch.cat([features_2d,features_2d.permute(0,1,3,2)],axis=1)


        left = features_1d.unsqueeze(2).repeat(1, 1, seqlen, 1)
        right = features_1d.unsqueeze(1).repeat(1, seqlen, 1, 1)
        features = torch.cat((left.permute(0,3,1,2), right.permute(0,3,1,2), features_2d), 1)
        features = features.contiguous()
        return features #[N,C,L,L]
    
class LMResnet(nn.Module):
    def __init__(self, lm_model, trrosetta_model):
        super().__init__()
        self.lm = lm_model.to(device) 
        self.extract_features_LM = LMFeatureExtractor().to(device) 
        self.trrosetta = trrosetta_model.to(device) 

    def forward(self,
                seq,
                seq1h,
                sel,
                idx_raw,
                msa,
                insertions,
                idx,
                theta=None,
                phi=None,
                dist=None,
                omega=None):
        
        LM_output = self.lm(seq, repr_layers=[33], return_contacts=True)

        features_LM = self.extract_features_LM(seq1h, LM_output)
        features_LM = features_LM[:,:,idx_raw,:]
        features_LM = features_LM[:,:,:,idx_raw]
        features_LM = features_LM[:,:,sel,:]
        features_LM = features_LM[:,:,:,sel]
    
        seqsep = torch.abs(idx[:,None]-idx[None,:])+1
        seqsep = torch.log(seqsep.float())
        
        features = torch.cat([features_LM, seqsep[None,None,:,:]], axis=1)
        
        return self.trrosetta(features, theta=theta, phi=phi, dist=dist, omega=omega)
    
#l2 regularization
def add_weight_decay(model, l2_value, skip_list=()):
    decay, no_decay = [], []
    for name, param in model.named_parameters():
        if not param.requires_grad: continue          
        if len(param.shape) == 1 or name.endswith(".bias") or name in skip_list: no_decay.append(param)
        else: decay.append(param)
    return [{'params': no_decay, 'weight_decay': 0.}, {'params': decay, 'weight_decay': l2_value}]